import os
import re
import subprocess
import sys
from subprocess import Popen

__all__ = ["CommandError", "run", "inject", "run_get"]

from dataclasses import dataclass
from textwrap import dedent
from typing import List, Optional, Union

from envo import console
from envo.misc import is_linux, is_windows


class CommandError(RuntimeError):
    pass


def _get_popen_cmd(command: str) -> List[str]:
    command_extra = re.sub(r"\\(?:\t| )*\n(?:\t| )*", "", command)

    if is_windows():
        command_extra = command_extra.strip()
        command_extra = command_extra.replace("\r", "")
        command_extra = command_extra.replace("\n", " & ")
        popen_cmd = ["cmd.exe", "/c", command_extra]
    elif is_linux():
        options = ["set -uoe pipefail", "shopt -s globstar"]
        command_extra = "\n".join(options) + "\n" + command_extra
        popen_cmd = ["/bin/bash", "--rcfile", "/dev/null", "-c", command_extra]
    else:
        raise NotImplementedError()

    return popen_cmd


@dataclass
class Output:
    stdout: str
    stderr: str


def run_get(command: str) -> Output:
    popen_cmds = _get_popen_cmd(command=command)

    proc = Popen(popen_cmds, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=os.environ)
    outs, errs = proc.communicate()

    output = Output(stdout=outs.decode("utf-8"), stderr=errs.decode("utf-8"))
    return output


def _run(
    command: str,
    raise_on_error=True,
    print_output=True,
    print_errors=True,
    verbose: Optional[bool] = None,
    background=False,
) -> Optional[str]:

    if verbose is None:
        verbose = os.environ.get("ENVO_VERBOSE_RUN", False)

    popen_cmd = _get_popen_cmd(command)

    if verbose:
        dedent_cmd = dedent(command.strip())
        console.rule(f"[bold rgb(225,221,0)]{dedent_cmd}", align="center", style="rgb(255,0,255)")

    kwargs = {}
    if not print_output:
        kwargs["stdout"] = subprocess.PIPE

    if not print_errors:
        kwargs["stderr"] = subprocess.PIPE

    proc = Popen(popen_cmd, env=os.environ, **kwargs)

    if background:
        return None

    ret_code = proc.wait()

    if ret_code != 0 and raise_on_error:
        sys.exit(ret_code)


def run(
    command: Union[str, List[str]],
    raise_on_error=True,
    print_output=True,
    print_errors=True,
    verbose: Optional[bool] = None,
    background=False,
    progress_bar: Optional[str] = None,
) -> None:
    # join multilines
    if not isinstance(command, list):
        command = [command]

    if progress_bar is not None:
        from rich.progress import track

        for c in track(command, description=progress_bar):
            _run(
                command=c,
                raise_on_error=raise_on_error,
                print_output=print_output,
                print_errors=print_errors,
                verbose=verbose,
                background=background,
            )
    else:
        for c in command:
            _run(
                command=c,
                raise_on_error=raise_on_error,
                print_output=print_output,
                print_errors=print_errors,
                verbose=verbose,
                background=background,
            )


def inject(command: str) -> None:
    __xonsh__.shell.run_code(command)



import inspect
import os
from copy import copy, deepcopy
from dataclasses import dataclass
from getpass import getpass
from typing import Any, Callable, List, Optional, TypeVar, cast

import envium.environ
from envium.environ import Var, VarGroup

from envo import CommandError, inject, run, run_get

__all__ = ["secret_var", "ctx_var", "Ctx", "SecretsGroup", "CtxGroup"]


VarType = TypeVar("VarType")


class CtxVar(Var[TypeVar]):
    pass


def ctx_var(
    default: Optional[VarType] = None,
    default_factory: Optional[Callable] = None,
) -> Any:
    return CtxVar(default=default, default_factory=default_factory)


class SecretVar(CtxVar):
    def __init__(
        self,
        default: Optional[VarType] = None,
        default_factory: Optional[Callable] = None,
        value_from_input: bool = True,
    ) -> None:
        super().__init__(default=default, default_factory=default_factory)
        self._value_from_input = value_from_input


def secret_var(default: Optional[VarType] = None, default_factory: Optional[Callable] = None) -> Any:
    return SecretVar(default=default, default_factory=default_factory)


class WrongTypeError(envium.environ.WrongTypeError):
    pass


class NoTypeError(envium.environ.NoTypeError):
    pass


class NoValueError(envium.environ.NoValueError):
    pass


class ValidationErrors(envium.environ.ValidationErrors):
    pass


class CtxGroup(VarGroup):
    pass


class Ctx(CtxGroup):
    def __init__(self, name: str) -> None:
        super().__init__(load=False, name=name)
        self._root = self
        self._process()

    @property
    def _fullname(self) -> str:
        return self._name

    @property
    def _errors(self) -> List[envium.environ.EnviumError]:
        envium_errors = super()._errors

        ret = []
        for e in envium_errors:
            if isinstance(e, envium.environ.NoValueError):
                ret.append(NoValueError(var_name=e.var_name, type_=e.type_))

            if isinstance(e, envium.environ.NoTypeError):
                ret.append(NoTypeError(var_name=e.var_name))

            if isinstance(e, envium.environ.WrongTypeError):
                ret.append(WrongTypeError(var_name=e.var_name, type_=e.type_, got_type=e.got_type))

        return ret

    def _validate(self) -> None:
        try:
            super()._validate()
        except envium.environ.ValidationErrors:
            raise ValidationErrors(self._errors)


class SecretsGroup(VarGroup):
    pass


class Secrets(Ctx):
    def __init__(self, name: str) -> None:
        super().__init__(name)

        self._get_secrets_from_input()

    @property
    def _flat(self) -> List[SecretVar]:
        return cast(List[SecretVar], super()._flat)

    def _get_secrets_from_input(self) -> None:
        for s in self._flat:
            if s._value_from_input:
                value = getpass(f"{s._fullname}: ")
                s._value = value



import os
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional, Union

import envo
from envo import Env, Namespace, logger

__all__ = [
    "Plugin",
    "VirtualEnv",
]

import envium
from envium import var

from envo.env import BaseEnv
from envo.logs import Logger
from envo.misc import is_windows


class Plugin(BaseEnv):
    class Environ:
        pass


class CantFindEnv(Exception):
    pass


@dataclass
class VenvPath:
    root_path: Path
    venv_name: str

    @property
    def path(self) -> Path:
        return self.root_path / self.venv_name

    @property
    def lib_path(self) -> Path:
        return self.path / "lib"

    @property
    def bin_path(self) -> Path:
        if is_windows():
            return self.path / "Scripts"
        else:
            return self.path / "bin"

    @property
    def site_packages_path(self) -> Path:
        if is_windows():
            ret = self.lib_path / "site-packages"
            return ret
        try:
            ret = next(self.lib_path.glob("*"))
            # on windows there is no python.* in site packages path
            if ret.name != "site-packages":
                ret /= "site-packages"
            return ret
        except StopIteration:
            raise CantFindEnv()

    @property
    def possible_site_packages(self) -> List[Path]:
        if is_windows():
            return [self.site_packages_path]
        versions = [f"python3.{v}" for v in range(0, 20)]
        versions += [f"python2.{v}" for v in range(0, 10)]
        ret = [self.lib_path / v / "site-packages" for v in versions]
        return ret


class BaseVenv:
    venv_path: VenvPath

    def __init__(self, root: Path, venv_dir_name: str) -> None:
        self.root = root
        self.venv_dir_name = venv_dir_name
        self._path_delimiter = ";" if is_windows() else ":"

    def _get_path(self):
        path = f"""{str(self.venv_path.bin_path)}"""
        return path

    def activate(self, e: Env.Environ) -> None:
        e.path = f"""{self._get_path()}{self._path_delimiter}{e.path}"""

    def deactivate(self, e: Env.Environ) -> None:
        if self._get_path() in e.path:
            e.path = e.path.replace(self._get_path() + self._path_delimiter, "")


class PredictedVenv(BaseVenv):
    def __init__(self, root: Path, venv_name: str) -> None:
        super().__init__(root, venv_name)

        self.venv_path = VenvPath(root_path=root, venv_name=venv_name)

    def activate(self, e: Env.Environ) -> None:
        super().activate(e)
        for d in self.venv_path.possible_site_packages:
            if str(d) not in sys.path:
                sys.path.insert(0, str(d))

    def deactivate(self, e: Env.Environ) -> None:
        super().deactivate(e)
        for d in self.venv_path.possible_site_packages:
            if str(d) in sys.path:
                sys.path.remove(str(d))


class ExistingVenv(BaseVenv):
    def __init__(self, root: Path, venv_name: str, discover: bool = False) -> None:
        super().__init__(root, venv_name)
        self.discover = discover

        root_path = self._get_venv_dir() if discover else root

        self.venv_path = VenvPath(root_path=root_path, venv_name=venv_name)

        if not self.venv_path.path.exists():
            raise CantFindEnv

    def _get_venv_dir(self) -> Optional[Path]:
        path = self.root

        while path.parent != path:
            if (path / self.venv_dir_name).exists():
                return path
            path = path.parent

        raise CantFindEnv()

    def activate(self, e: Env.Environ) -> None:
        super().activate(e)

        if str(self.venv_path.site_packages_path) not in sys.path:
            sys.path.insert(0, str(self.venv_path.site_packages_path))

    def deactivate(self, e: Env.Environ) -> None:
        super().deactivate(e)

        try:
            while str(self.venv_path.site_packages_path) in sys.path:
                sys.path.remove(str(self.venv_path.site_packages_path))
        except CantFindEnv:
            pass


class VirtualEnv(Plugin):
    """
    Env that activates virtual environment.
    """

    class Environ(envium.Environ):
        path: str = var(raw=True)

    e: Environ

    @classmethod
    def customise(cls, venv_path: Optional[Union[Path, str]] = None, venv_name: str = ".venv"):
        cls.venv_path = Path(venv_path).absolute() if venv_path else None
        cls.venv_name = venv_name

    def init(self):
        super().init()

        root = Path(".").absolute()
        # Handle standalone and inherited case
        if hasattr(self, "e"):
            e = self.e
        else:
            e = VirtualEnv.Environ(name="envo", load=True)
        self.__logger: Logger = logger.create_child("venv", descriptor="VirtualEnv")

        venv_path = root if not self.venv_path else self.venv_path

        self._possible_site_packages = []
        self._venv_dir_name = self.venv_name

        self.__logger.debug("VirtualEnv plugin init")

        try:
            self._venv = ExistingVenv(
                root=venv_path,
                venv_name=self.venv_name,
                discover=self.venv_path is None,
            )

        except CantFindEnv:
            self.__logger.debug("Couldn't find venv. Falling back to predicting")
            self._venv = PredictedVenv(root=venv_path, venv_name=self.venv_name)

        self._venv.activate(e)
        os.environ["PATH"] = e.path


VirtualEnv.customise()
